/*
 * Copyright (c) 2007, 2008 ICM Uniwersytet Warszawski All rights reserved.
 * See LICENCE file for licencing information.
 *
 * Created on Mar 17, 2007
 * Author: K. Benedyczak <golbi@mat.umk.pl>
 */

package pl.edu.icm.unicore.uvos.db;

import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.Reader;
import java.sql.Connection;
import java.sql.SQLException;
import java.util.Calendar;
import java.util.Date;
import java.util.HashMap;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Properties;
import java.util.Vector;
import java.util.Map.Entry;

import org.apache.ibatis.exceptions.PersistenceException;
import org.apache.ibatis.executor.ExecutorException;
import org.apache.ibatis.session.SqlSession;
import org.apache.ibatis.session.SqlSessionFactory;
import org.apache.ibatis.session.SqlSessionFactoryBuilder;
import org.apache.log4j.Logger;

import pl.edu.icm.unicore.uvos.api.Attribute;
import pl.edu.icm.unicore.uvos.api.exceptions.InternalException;
import pl.edu.icm.unicore.uvos.util.ConfDefaults;
import pl.edu.icm.unicore.uvos.util.Configuration;

/**
 * Provides general DB handling methods and support for obtaining MyBatis SqlSessions.
 * 
 * @author K. Benedyczak
 */
public class DB
{
	private static final Logger log = Logger.getLogger(DB.class);

	public static final String DB_VERSION = "1_2_0";
	
	//all properties from the datamap.properties file 
	//(defined in the main configuration file)
	public static final String P_URL = "url";
	public static final String P_DRIVER = "driver";
	public static final String P_USERNAME = "username";
	public static final String P_PASSWD = "password";

	public static final String INIT_DATA_SCRIPT_DEF = "conf/db/initdata.sql";
	public static final String CLEAR_SCRIPT_DEF = "conf/db/cleardb.sql";
	public static final String CHECK_SCRIPT_DEF = "conf/db/checkdb.sql";

	public static final String P_INIT_DATA_SCRIPT = "initdata.script";
	public static final String P_CLEAR_SCRIPT = "cleardb.script";
	public static final String P_CHECK_SCRIPT = "checkdb.script";
	public static final String P_CREATEDB_SCRIPT = "createdb.script";
	
	private SqlSessionFactory sqlMapFactory;
	private boolean shutdown = false;
	private DBLock lock;

	
	public DB(Configuration config, DBLock lock) throws InternalException, FileNotFoundException
	{
		this.lock = lock;
		String configFile = config.getAsString(ConfDefaults.P_DBCONFIG_FILE, ConfDefaults.DBCONFIG_FILE_DEF, log);
		Reader reader = new FileReader(configFile);
		SqlSessionFactoryBuilder builder = new SqlSessionFactoryBuilder();
		sqlMapFactory = builder.build(reader);
		verifyDBVersion(configFile);
	}
	
	public static List<Attribute> cleanupAttributeList(List<Attribute> withDuplicates) 
	{
		HashMap<String, Attribute> hash = new HashMap<String, Attribute>();  
		for (Attribute input: withDuplicates)
		{
			String typeName = input.getName().toString();
			String scope = input.getScope();
			typeName = (scope == null ? "" : scope) + "//" + typeName;
			Attribute at = hash.get(typeName);
			if (at == null)
				hash.put(typeName, input);
			else
			{
				List<String> newVals = input.getValues();
				if (newVals != null)
				{
					List<String> curVals = at.getValues();
					for (String newVal: newVals)
						if (!curVals.contains(newVal))
							curVals.add(newVal);
				}
			}
		}
		Vector<Attribute> ret = new Vector<Attribute>();
		Iterator<Attribute> hashVals = hash.values().iterator();
		while (hashVals.hasNext())
			ret.add(hashVals.next());
		return ret;		
	}
	
	private void verifyDBVersion(String configFile) throws InternalException
	{
		String actualDbVersion;
		try
		{
			actualDbVersion = checkCurrentVersion();
		} catch (Exception e)
		{
			throw new InternalException("Can't read version of the database. " +
				"Have you initialized it? Are connection details correctly " +
				"entered in configuration? The error was:\n\n" + e);
		}
		if (!actualDbVersion.equals(DB.DB_VERSION))
			throw new InternalException("The database is initialized with " +
				"wrong schema. It is of UVOS version: " + actualDbVersion + 
				" while you are using now UVOS version:" + DB.DB_VERSION);
	}
	
	private static class Holder
	{
		private long timestamp;
		private StackTraceElement[] stackTrace;
		public Holder(long timestamp, StackTraceElement[] stackTrace)
		{
			this.timestamp = timestamp;
			this.stackTrace = stackTrace;
		}
	}
	private Map<SqlSession, Holder> used = new HashMap<SqlSession, Holder>();
	
	public synchronized SqlSession getSqlSession()
	{
		long now = System.currentTimeMillis();
		Holder h = new Holder(now, Thread.currentThread().getStackTrace());
		
		SqlSession newSession = sqlMapFactory.openSession();
		Holder p = used.put(newSession, h);
		if (p != null)
			log.warn("Ups! MyBatis returned a SqlSession which is already used!");
		Iterator<Entry<SqlSession, Holder>> it = used.entrySet().iterator();
		while(it.hasNext())
		{
			Entry<SqlSession, Holder> e = it.next();
			long wait = now - e.getValue().timestamp;
			if (wait > 2000)
			{
				log.warn("SqlSession is kept for more than 1s: " + now/1000.0 + "s");
				StackTraceElement[] stackTrace = e.getValue().stackTrace;
				for (StackTraceElement se: stackTrace)
					log.warn("  " + se.toString());
			}
		}
		return newSession;
	}
	
	public synchronized void releaseSqlSession(SqlSession session)
	{
		Holder p = used.remove(session);
		if (p == null)
		{
			log.warn("Thread trying to release not known session: ");
			Thread.dumpStack();
		}
		session.close();
	}

	public void shutdown()
	{
		if (shutdown)
			return;
		SqlSession sqlMap = getSqlSession();
		try
		{
			Properties dbConf = sqlMap.getConfiguration().getVariables();
			String driver = dbConf.getProperty("driver");
			if ("org.hsqldb.jdbcDriver".equals(driver))
			{
				Connection conn = sqlMap.getConnection();
				conn.createStatement().execute("SHUTDOWN");
			}
			shutdown = true;
		} catch (ExecutorException e) 
		{
			//OK - means that sqlMap was already closed
		} catch (SQLException e)
		{
			log.warn("Can't close DB connection", e);
		} catch (PersistenceException e)
		{
			log.warn("Can't close DB connection", e);
		} finally
		{
			releaseSqlSession(sqlMap);
		}
	}
	
	public String checkCurrentVersion() throws Exception
	{
		SqlSession sqlMap = getSqlSession();
		try
		{
			List<?> versions = sqlMap.selectList("getDBVersions");
			String ver = "1_0_0";
			for (Object o: versions)
			{
				String v = (String)o;
				v = v.trim();
				if (ver.compareTo(v) < 0)
					ver = v;
			}
			return ver;
		} finally
		{
			releaseSqlSession(sqlMap);
		}
	}
	
	public void purgeHistory(Date date, SqlSession sqlMap) throws InternalException
	{
		lock.writeLock();
		try
		{
			sqlMap.delete("purgeHistory1", date);
			sqlMap.delete("purgeHistory2", date);
			sqlMap.delete("purgeHistory3", date);
			sqlMap.delete("purgeHistory4", date);
			sqlMap.delete("purgeHistory5", date);
			sqlMap.delete("purgeHistory6", date);
			sqlMap.delete("purgeHistory7", date);
		} catch (PersistenceException e)
		{
			String msg = "Error purging history: " + e.getMessage();
			log.error(msg + ". SQL exception is: " + e);
			throw new InternalException(msg);  
		}
		finally
		{
			lock.writeUnlock();
		}
	}
}
